#pragma once

#include "MainForm.rh"
#include "GUIThreads.h"
#include "Controls.h"
#include "StdControls.h"
#include "Dialogs.h"
#include "GUILog.h"
#include "VFingerX.h"
#include "ScanManX.h"
#include "VFFileList.h"
#include "VFImage.h"
#include "VFFeatures.h"
#include "VFDatabase.h"
#include "VFControls.h"
#include "VFOptions.h"
#include "Options.h"

enum VFDemoMode {vfdmNone = -1, vfdmEnrollment, vfdmEnrollmentWithGen, vfdmVerification, vfdmIdentification};

string VFDemoModeToInvariantString(VFDemoMode value);
VFDemoMode VFDemoModeFromInvariantString(const string &value);
string VFDemoModeToString(VFDemoMode value);

/////////////////
//             //
// CTaskThread //
//             //
/////////////////

const UINT UM_TASK_STATE = CM_USER;
const UINT UM_TASK_PROGRESS = CM_USER + 1;

class CTaskThread: public CGUIThread
{
private:
	string task_name;
	string task_string;

protected:
	bool failed;

	void DoTaskState(const string &state)
	{
		Control->SendMessage(UM_TASK_STATE, NULL, (LPARAM)(LPCTSTR)state);
	}
	bool DoTaskProgress(bool started, int value)
	{
		return Control->SendMessage(UM_TASK_PROGRESS, started, value) != 0;
	}

public:
	CTaskThread(CControl * control, const string &task_name, const string &task_string)
		: CGUIThread(control), task_name(task_name), task_string(task_string), failed(false)
	{
	}

	virtual ~CTaskThread()
	{
	}

	const string & TaskName() { return task_name; }
	const string & TaskString() { return task_string; }

	bool Failed() { return failed; }

};

///////////////
//           //
// CDBThread //
//           //
///////////////

class CDBThread: public CTaskThread
{
private:
	exception error;

protected:
	CVFDatabase *database;

	DWORD OnExecute()
	{
		try
		{
			CoInitialize(NULL);
			OnDBExecute();
			CoUninitialize();
		}
		catch(exception &e)
		{
			error = e;
			failed = true;
		}
		return failed;
	}

	virtual void OnDBExecute() = 0;

public:
	CDBThread(CControl * control, CVFDatabase *database, const string &operation_name, const string &operation_string)
		: CTaskThread(control, operation_name, operation_string)
	{
		this->database = database;
	}

	CVFDatabase* Database() { return database; }
	void SetDatabase(CVFDatabase *value) { CheckRunning(); database = value; }

	exception& Error() { return error; }
};

///////////////////
//               //
// CDBOpenThread //
//               //
///////////////////

class CDBOpenThread: public CDBThread, IVFDBOpenCallback
{
private:
	string provider;
	string data_source;
	string user_id;
	string password;

	void OnOpenState(const string &state)
	{
		DoTaskState(state);
	}

	bool OnOpenProgress(bool started, int value)
	{
		return DoTaskProgress(started, value);
	}


protected:
	void OnDBExecute() { database->Open(provider, data_source, user_id, password, this); }

public:
	CDBOpenThread(CControl * control = NULL, CVFDatabase *database = NULL,
		const string &provider = emptyString, const string &data_source = emptyString,
		const string &user_id = emptyString, const string &password = emptyString)
		: CDBThread(control, database, STRING("Database open"), STRING("Opening database"))
	{
		this->database = database;
		this->provider = provider;
		this->data_source = data_source;
		this->user_id = user_id;
		this->password = password;
	}

};

////////////////////
//                //
// CDBCloseThread //
//                //
////////////////////

class CDBCloseThread: public CDBThread
{
protected:
	void OnDBExecute() { database->Close(); }

public:
	CDBCloseThread(CControl * control = NULL, CVFDatabase *database = NULL)
		: CDBThread(control, database, STRING("Database close"), STRING("Closing database"))
	{
	}
};

////////////////////
//                //
// CDBClearThread //
//                //
////////////////////

class CDBClearThread: public CDBThread
{
protected:
	void OnDBExecute() { database->Clear(); }

public:
	CDBClearThread(CControl * control = NULL, CVFDatabase *database = NULL)
		: CDBThread(control, database, STRING("Database clear"), STRING("Clearing database"))
	{
	}
};

/////////////////////
//                 //
// CFileListThread //
//                 //
/////////////////////

class CFileListThread: public CTaskThread
{
private:
	VFFileList *file_list;

protected:
	DWORD OnExecute()
	{
		DoTaskProgress(false, file_list->Length());
		for(int i = 0; i < file_list->Length(); i++)
			if(!DoTaskProgress(true, i)) break;
		return 0;
	}

public:
	CFileListThread(CControl * control, VFFileList *file_list)
		: CTaskThread(control, STRING("File list"), STRING("Processing file list"))
	{
		this->file_list = file_list;
	}

};

///////////////////////
//                   //
// CExtractionThread //
//                   //
///////////////////////

class CExtractionThread: public CGUIThread
{
private:
	int resolution;
	CVFImage *image;
	CVFImage *image_after;

	BYTE features[VF_MAX_FEATURES_SIZE];
	DWORD features_size;

	DWORD time;

	INT result;

protected:
	DWORD OnExecute()
	{
		image_after->Assign(image);
		DWORD tc = GetTickCount();
		result = VFExtract(image->Width(), image->Height(), image_after->Bits(), resolution, features, &features_size, NULL);
		time = GetTickCount() - tc;
		return (DWORD)result;
	}

public:
	CExtractionThread(CControl *control, int resolution, CVFImage *image, CVFImage *image_after)
		: CGUIThread(control), result(-1), time(0), resolution(resolution), image(image), image_after(image_after)
	{
	}

	BYTE *Features() { return features; }
	DWORD FeaturesSize() { return features_size; }
	DWORD Time() { return time; }
	INT Result() { return result; }

};

///////////////////////////
//                       //
// CIdentificationThread //
//                       //
///////////////////////////

const UINT UM_IDENTIFICATION_PROGRESS = CM_USER + 3;

const int IDEN_PROGRESS_COUNT = 1000;

struct IdenProgress
{
	int count;
	int index[IDEN_PROGRESS_COUNT];
	INT result[IDEN_PROGRESS_COUNT];
	int similarity[IDEN_PROGRESS_COUNT];
};

class CIdentificationThread: public CGUIThread
{
private:
	BYTE * features;
	DWORD featuresSize;
	CVFDatabase *database;
	bool use_g;
	int database_percent;
	bool match_all;
	INT result;
	int index;
	int similarity;
	double time;

protected:
	DWORD OnExecute()
	{
		result = VFIdentifyStart(features, NULL);
		if(VFSucceeded(result))
		{
			VFDBFindData fd;
			if(use_g) database->FindStart(features, fd);
			int c = database->Count;
			int count_threshold = c * database_percent / 100, count = 0;
			Control->SendMessage(UM_IDENTIFICATION_PROGRESS, FALSE, c);
			IdenProgress ip;
			ip.count = 0;
			VFMatchDetails md;
			md.Size = sizeof(md);
			bool stop = false;
			int i = -1;
			DWORD dt = GetTickCount();
			do
			{
				if(use_g)
				{
					i = database->FindNext(fd);
					if(i == -1 || count > count_threshold) i = c;
				}
				else i++;
				if(i < c)
				{
					time -= GetTickCountEx();
					result = VFIdentifyNext(database->Features(i), &md, NULL);
					time += GetTickCountEx();
					ip.index[ip.count] = i;
					ip.result[ip.count] = result;
					ip.similarity[ip.count] = md.Similarity;
					ip.count++;
					if(VFSucceeded(result))
					{
						if(index == -1 || md.Similarity > similarity)
						{
							index = i;
							similarity = md.Similarity;
							stop = !match_all;
						}
					}
					count++;
				}
				else stop = true;
				if(ip.count == IDEN_PROGRESS_COUNT || stop || GetTickCount() - dt > 300)
				{
					if(!Control->SendMessage(UM_IDENTIFICATION_PROGRESS, TRUE, (LPARAM)&ip)) break;
					ip.count = 0;
					dt = GetTickCount();
				}
			} while(!stop);
			result = VFIdentifyEnd(NULL);
			if(use_g) database->FindEnd(fd);
			if(VFSucceeded(result)) result = index == -1 ? VFE_FAILED : VFE_OK;
		}
		return (DWORD)result;
	}

public:
	CIdentificationThread(CControl * control, const BYTE * features, DWORD featuresSize, CVFDatabase *database, bool use_g, int database_percent, bool match_all)
		: CGUIThread(control), use_g(use_g), database_percent(database_percent), match_all(match_all), result(VFE_FAILED), index(-1), similarity(0), time(0)
	{
		this->features = new BYTE[featuresSize];
		this->featuresSize = featuresSize;
		memcpy(this->features, features, featuresSize);
		this->database = database;
	}

	virtual ~CIdentificationThread()
	{
		delete[] features;
	}

	INT Result() { return result; }
	int Index() { return index; }
	int Similarity() { return similarity; }
	double Time() { return time; }

};

///////////////
//           //
// CMainForm //
//           //
///////////////

class CMainForm: public CForm
{
private:
	CVFDatabase * database;

	CTaskThread * task_thread;
	CExtractionThread * extraction_thread;
	CIdentificationThread * identification_thread;
	bool stop_task, stop_identification;

	int resolution;
	CVFImage * image;
	CVFImage * image_after;
	CVFFeatures * features;
	string finger_id;
	bool image_from_scanner;
	string scanner_id;

	int gen_index;
	int gen_resolution[VF_GENERALIZE_COUNT];
	CVFImage * gen_image[VF_GENERALIZE_COUNT];
	CVFImage * gen_image_after[VF_GENERALIZE_COUNT];
	CVFFeatures * gen_features[VF_GENERALIZE_COUNT];
	string gen_finger_id[VF_GENERALIZE_COUNT];

	int ver_index;
	int ver_resolution[2];
	CVFImage * ver_image[2];
	CVFImage * ver_image_after[2];
	CVFFeatures * ver_features[2];
	string ver_finger_id[2];

	string appName;
	string engineName;
	
	VFDemoMode mode;
	VFDemoMode lastMode;
	bool changeBackMode;

	string databaseFileName;
	bool createDatabase;

	string settingsFileName;

	VFFileList * file_list;

	VFOptions vfOptions;
	VFDemoOptions options;

	DWORD task_time_started;
	DWORD task_ui_time;
	double identification_time_started;

	bool finger_found;
	int count, count_rej;
	int count_iden, total_iden;
	DWORD total_size;
	DWORD total_ext_time;
	double total_iden_time;
	int fingers_found;
	int false_accepted, false_rejected;

	CVFView * view_left, * view_right;
	CVSplitter * splitter_views;
	CHSplitter * splitter_horz;
	CCaptionPanel * captionpanel_main, * captionpanel_scanner, * captionpanel_results;
	CPanel * panel_main, * panel_results;
	CLabel * label_main;
	CProgressBar * progressbar_main, * progressbar_results;
	CLabel * label_main_time, * label_results_time;
	CButton * button_main_stop, * button_results_stop;
	CRichEdit * richedit_main;
	CRichEdit * richedit_scanner;
	CListView * listview_results;
	CVSplitter * splitter_v1, * splitter_v2;
	CStatusBar * statusbar;

	COpenFileDialog * open_dialog;
	COpenFileDialog * open_list_dialog;
	CSaveFileDialog * save_dialog;

	CGUILog * main_log, * scanner_log;

	void InitControls();
	void AdjustControls(bool top_part, bool bottom_part);

	void LoadSettings();
	void SaveSettings();

	void LoadSettings(MSXML2::IXMLDOMDocumentPtr document);
	void SaveSettings(MSXML2::IXMLDOMDocumentPtr document);

	string filesDirectory;
	string fileListsDirectory;

	void LoadOptions(MSXML2::IXMLDOMNodePtr node);
	void SaveOptions(MSXML2::IXMLDOMDocumentPtr document, MSXML2::IXMLDOMNodePtr node);

	void AddVersionToLog(CLog *log, const string &name, DWORD version_high, DWORD version_low);
	string FingerprintCountToString(int value, const string &no = emptyString)
	{
		return IntToString(value, STRING("fingerprint"), STRING("fingerprints"), no);
	}
	string MatchingSpeedToString(int count, double time, const string &unknown = STRING("unknown"))
	{
		return time ? FingerprintCountToString(Round(count / time)) + TEXT(" per second") : unknown;
	}
	string PercentToString(int count, int total, const string &unknown = STRING("Unknown"))
	{
		return total ? Format(TEXT("%d%%"), Percent(count, total)) : unknown;
	}

	string GetModeString();
	void UpdateMode();

	void UpdateStatus();

	void StartTask(CTaskThread *thread);
	void OnTaskState(const string &state);
	bool OnTaskProgress(bool started, int value);
	void OnTaskComplete();
	void ChangeBackMode();
	string GetTaskName();
	string GetTaskString();
	void UpdateTask();
	void UpdateImagesMenu();
	void UpdateTaskProgress(bool started, int value, const string &text = emptyString);
	void UpdateIdentificationProgress(bool started, int value);
	bool TaskRunning();
	bool CanStopTask();
	bool StoppingTask();
	bool MainTaskRunning();
	bool DBOpenRunning();
	bool FileListRunning();
	bool IdentificationRunning();
	bool StoppingIdentification();
	void Clear();
	void StopTask();
	void StopIdentification();
	bool StopTaskPrompt();
	bool ErrorPrompt(const string &reason, bool warning = false);

	void CreateDB(const string &filename);
	void StartDBOpen();
	void OnDBOpenComplete(DWORD time);
	void CloseDB();
	void ReOpenDB();

	void Enroll();
	void EnrollWithGen();
	void Verify();
	bool StartIdentification();

	void OnFileOpen();
	void OnFileOpenFileList();
	void OnFileSave();
	void OnFileSaveRight();
	void OnFileStop();
	void OnFileExit();
	void StartFileList();
	void OnFileListImage(int file_index);
	void OnFileListComplete(DWORD time);
	bool OnFileImage(const string &filename, int resolution, const string &finger_id);
	
	static void SM_CALLBACK SMMonitorProc(const CHAR * scanner_id, DWORD event, VOID * param)
	{
		((CMainForm*)param)->OnSMMonitor(AnsiToString(scanner_id), event);
	}
	static void SM_CALLBACK SMImageProc(const CHAR * scanner_id, INT width, INT height, const BYTE * image, INT resolution, VOID * param)
	{
		((CMainForm*)param)->OnScannerImage(AnsiToString(scanner_id), width, height, image, resolution);
	}
	static void SM_CALLBACK SMStateProc(const CHAR * scanner_id, DWORD state, VOID * param)
	{
		((CMainForm*)param)->OnScannerState(AnsiToString(scanner_id), state);
	}
	void OnSMMonitor(const string & scanner_id, DWORD event);
	void OnScannerImage(const string & scanner_id, int width, int height, const BYTE * image, int resolution);
	void OnScannerState(const string & scanner_id, DWORD state);

	void OnImage();
	void OnExtractionComplete();
	bool OnIdentificationProgress(bool started, IdenProgress *progress);
	void OnIdentificationComplete();
	void SaveImage(CVFImage *image);

	void OnModeEnrollment();
	void OnModeEnrollmentWithGen();
	void OnModeIdentification();
	void OnModeVerification();

	void OnViewZoomIn();
	void OnViewZoomOut();
	void OnViewZoom1To1();
	void OnViewShowFeatures();
	void OnViewShowSP();
	void OnViewShowBO();
	void UpdateViewMenu();

	void OnToolsFileListEditor();
	void OnToolsClearDatabase();
	void OnToolsClearMainLog();
	void OnToolsClearScannerLog();
	void OnToolsOptions();
	void OnToolsEngineOptions();

	void OnHelpAbout();

protected:
	LPCTSTR GetMenuName() { return MAKEINTRESOURCE(IDM_MAIN); }
	LPCTSTR GetAccelName() { return MAKEINTRESOURCE(IDA_MAIN); }
	HICON GetIcon() { return Application()->LoadIcon(MAKEINTRESOURCE(IDI_VFDEMO)); }
	HICON GetSmallIcon() { return Application()->LoadIcon(MAKEINTRESOURCE(IDI_VFDEMO_SMALL)); }

	LRESULT OnMessage(UINT message, WPARAM wParam, LPARAM lParam);
	
	void OnCreate();
	void OnDestroy();
	LRESULT OnSize(DWORD resizing_type, WORD client_width, WORD client_height);
	void OnCommand(WORD id, WORD event, CControl *control);
	void OnFirstShow();
	bool OnCloseQuery();
	void OnClose();

public:
	CMainForm(const string &appName, const string &engineName);
	~CMainForm();

	const string & get_FilesDirectory() { return filesDirectory; }
	void set_FilesDirectory(const string & value) { filesDirectory = value; }
	__declspec(property(get = get_FilesDirectory, put = set_FilesDirectory)) const string & FilesDirectory;

	const string & get_FileListsDirectory() { return fileListsDirectory; }
	void set_FileListsDirectory(const string & value) { fileListsDirectory = value; }
	__declspec(property(get = get_FileListsDirectory, put = set_FileListsDirectory)) const string & FileListsDirectory;

	VFDemoMode get_Mode() { return mode; }
	void set_Mode(VFDemoMode value)
	{
		if(!TaskRunning() && mode != value)
		{
			mode = value;
			main_log->WriteLine(TEXT("Switching to mode: %s"), (LPCTSTR)GetModeString());
			main_log->NewLine();
			Clear();
			UpdateMode();
		}
	}
	__declspec(property(get = get_Mode, put = set_Mode)) VFDemoMode Mode;

	const string & get_DatabaseFileName() { return databaseFileName; }
	void set_DatabaseFileName(const string & value) { SetDatabaseFileName(value, createDatabase); }
	__declspec(property(get = get_DatabaseFileName, put = set_DatabaseFileName)) const string & DatabaseFileName;

	bool get_CreateDatabase() { return createDatabase; }
	void set_CreateDatabase(bool value) { if(createDatabase != value) SetDatabaseFileName(databaseFileName, value); }
	__declspec(property(get = get_CreateDatabase, put = set_CreateDatabase)) bool CreateDatabase;

	void SetDatabaseFileName(const string &value, bool createDatabase)
	{
		this->createDatabase = createDatabase;
		string oldDatabaseFileName = GetFullFileName(databaseFileName);
		bool oldDB = !GetFileName(oldDatabaseFileName).IsEmpty();
		string newDatabaseFileName = GetFullFileName(value);
		bool newDB = !GetFileName(newDatabaseFileName).IsEmpty();
		bool otherDatabase = !oldDatabaseFileName.Equals(newDatabaseFileName, true);
		if(otherDatabase) databaseFileName = value;
		if((otherDatabase && (oldDB || newDB)) || (!otherDatabase && createDatabase && !database->Opened() && newDB))
			ReOpenDB();
	}
};
