#include "System.h"
#include "VFControls.h"
#include "VFingerX.h"
#include "Utils.h"
#include <math.h>

const int CVFView::MinZoom = 25;
const int CVFView::MaxZoom = 800;

CVFView::CVFView()
	: CControl(), image(NULL), maxColor(RGB(255, 255, 255)), minColor(RGB(0, 0, 0)),
		bitmap(NULL), bitmapWidth(0), bitmapHeight(0), resolution(VF_IMAGE_RESOLUTION),
		features(NULL), showFeatures(true), minutiaColor(RGB(255, 0, 0)),
		showSingularPoints(true), singularPointColor(RGB(255, 127, 0)),
		showBlockedOrientations(false), blockedOrientationsColor(RGB(0, 0, 255)),
		zoom(100)
{
	SetClientEdge(true);
	SetSize(300, 300);
}

CVFView::~CVFView()
{
	FreeBitmap();
}

void CVFView::FreeBitmap()
{
	if(bitmap)
	{
		DeleteBitmap(bitmap);
		bitmap = NULL;
		bitmapWidth = 0;
		bitmapHeight = 0;
	}
}

void CVFView::UpdateImage()
{
	FreeBitmap();
	if(image)
	{
		bitmap = image->GetBitmap(minColor, maxColor);
		bitmapWidth = image->Width();
		bitmapHeight = image->Height();
	}
	Invalidate();
}

void CVFView::UpdateFeatures()
{
	Invalidate();
}

void CVFView::GetImageRect(int &x, int &y, int &w, int &h)
{
	w = (bitmap ? bitmapWidth * VF_IMAGE_RESOLUTION / resolution : 0) * zoom / 100;
	h = (bitmap ? bitmapHeight * VF_IMAGE_RESOLUTION / resolution : 0) * zoom / 100;
	x = (ClientWidth() - w) / 2;
	y = (ClientHeight() - h) / 2;
}

LRESULT CVFView::OnEraseBackground(HDC dc, LPARAM /*lParam*/)
{
	RECT r;
	HBRUSH br = BackgroundBrush();
	if(image)
	{
		int x, y, w, h;
		GetImageRect(x, y, w, h);
		r.left = 0;
		r.top = 0;
		r.right = ClientWidth();
		r.bottom = y;
		FillRect(dc, &r, br);
		r.top = r.bottom;
		r.bottom += h;
		r.right = x;
		FillRect(dc, &r, br);
		r.left = r.right + w;
		r.right = ClientWidth();
		FillRect(dc, &r, br);
		r.left = 0;
		r.top = r.bottom;
		r.bottom = ClientHeight();
		FillRect(dc, &r, br);
	}
	else
	{
		GetClientRect(r);
		FillRect(dc, &r, br);
	}
	return TRUE;
}

LRESULT CVFView::OnPaint(WPARAM /*wParam*/, LPARAM /*lParam*/)
{
	PAINTSTRUCT ps;
	HDC dc = BeginPaint(Handle(), &ps);
	int x, y, w, h;
	GetImageRect(x, y, w, h);
	if(bitmap)
	{
		HDC bmp_dc = CreateCompatibleDC(dc);
		HBITMAP old_bitmap = SelectBitmap(bmp_dc, bitmap);
		StretchBlt(dc, x, y, w, h, bmp_dc, 0, 0, bitmapWidth, bitmapHeight, SRCCOPY);
		SelectBitmap(bmp_dc, old_bitmap);
		DeleteDC(bmp_dc);
	}
	if(showFeatures && features)
	{
		HBRUSH old_brush = SelectBrush(dc, GetStockBrush(NULL_BRUSH));
		if(showBlockedOrientations)
		{
			HPEN pen = CreatePen(PS_SOLID, max(1, Round(FeatToLocalF(1.0))), blockedOrientationsColor);
			HPEN old_pen = SelectPen(dc, pen);
			int bs = FeatToLocal(VF_BLOCK_SIZE);
			int hbs = bs / 2;
			for(int i = 0, yy = y + hbs; i < features->BlockedOrientations.Height(); i++, yy += bs)
				for(int j = 0, xx = x + hbs; j < features->BlockedOrientations.Width(); j++, xx += bs)
				{
					BYTE dir = features->BlockedOrientations.GetPixel(j, i);
					if(dir != VFDIR_BACKGROUND)
					{
						if(dir != VFDIR_UNKNOWN) PaintBO(dc, xx, yy, dir);
						else
						{
							PaintBO(dc, xx, yy, VFDIR_0);
							PaintBO(dc, xx, yy, VFDIR_45);
							PaintBO(dc, xx, yy, VFDIR_90);
							PaintBO(dc, xx, yy, VFDIR_135);
						}
					}
				}
			SelectPen(dc, old_pen);
			DeletePen(pen);
		}
		if(showSingularPoints)
		{
			HPEN pen = CreatePen(PS_SOLID, max(1, Round(FeatToLocalF(1.0))), singularPointColor);
			HPEN old_pen = SelectPen(dc, pen);
			int xx, yy, s = max(1, Round(FeatToLocalF(5.0)));
			for(int i = 0; i < features->SingularPoints.Length(); i++)
			{
				xx = x + FeatToLocal(features->SingularPoints[i].X);
				yy = y + FeatToLocal(features->SingularPoints[i].Y);
				switch(features->SingularPoints[i].T)
				{
					case vfsptCore:
						Rectangle(dc, xx - s, yy - s, xx + s + 1, yy + s + 1);
						if(features->SingularPoints[i].D != VFDIR_UNKNOWN)
							PaintDir(dc, xx, yy, features->SingularPoints[i].D, 20);
						break;
					case vfsptDoubleCore:
						Ellipse(dc, xx - s, yy - s, xx + s + 1, yy + s + 1);
						break;
					case vfsptDelta:
						MoveToEx(dc, xx - s, yy + s, NULL);
						LineTo(dc, xx, yy - s);
						LineTo(dc, xx + s, yy + s);
						LineTo(dc, xx - s, yy + s);
						break;
				}
			}
			SelectPen(dc, old_pen);
			DeletePen(pen);
		}
		HPEN pen = CreatePen(PS_SOLID, max(1, Round(FeatToLocalF(1.0))), minutiaColor);
		HPEN old_pen = SelectPen(dc, pen);
		int xx, yy, s = max(1, Round(FeatToLocalF(3.0)));
		for(int i = 0; i < features->Minutiae.Length(); i++)
		{
			xx = x + FeatToLocal(features->Minutiae[i].X);
			yy = y + FeatToLocal(features->Minutiae[i].Y);
			Ellipse(dc, xx - s, yy - s, xx + s + 1, yy + s + 1);
			PaintDir(dc, xx, yy, features->Minutiae[i].D, 10);
		}
		SelectPen(dc, old_pen);
		DeletePen(pen);
		SelectBrush(dc, old_brush);
	}
	EndPaint(Handle(), &ps);
	return 0;
}

void CVFView::PaintBO(HDC dc, int x, int y, BYTE dir)
{
	double hbs = FeatToLocalF(VF_BLOCK_SIZE / 2.0);
	double a = VFDirToRad(dir);
	int dx = Round(cos(a) * hbs);
	int dy = Round(sin(a) * hbs);
	MoveToEx(dc, x, y, NULL);
	LineTo(dc, x + dx, y + dy);
	MoveToEx(dc, x, y, NULL);
	LineTo(dc, x - dx, y - dy);
}

void CVFView::PaintDir(HDC dc, int x, int y, BYTE dir, int length)
{
	double len = FeatToLocalF(length);
	double a = VFDirToRad(dir);
	int dx = Round(cos(a) * len);
	int dy = Round(sin(a) * len);
	MoveToEx(dc, x, y, NULL);
	LineTo(dc, x + dx, y + dy);
}

void CVFView::ZoomIn()
{
	Zoom *= 2;
}

void CVFView::ZoomOut()
{
	Zoom /= 2;
}
