#include "System.h"
#include "Utils.h"
#include "VFDatabase.h"

CVFDatabase::CVFDatabase()
	: connected(false), opened(false), recordsetName(TEXT("Fingerprints")),
	connection(NULL), recordset(NULL), recordBinding(NULL)
{
	for(int g = 0; g < 256; g++)
	{
		items[g] = NULL;
		capacity[g] = 0;
		count[g] = 0;
	}
	allItems = NULL;
	allCount = 0;
	allCapacity = 0;
}

CVFDatabase::~CVFDatabase()
{
	Close();
}

int CVFDatabase::ID(int index)
{
	CheckIndex(index);
	return allItems[index]->id;
}

const BYTE * CVFDatabase::Features(int index)
{
	CheckIndex(index);
	return allItems[index]->features;
}

DWORD CVFDatabase::FeaturesSize(int index)
{
	CheckIndex(index);
	return allItems[index]->featuresSize;
}

string CVFDatabase::FingerID(int index)
{
	CheckIndex(index);
	return allItems[index]->fingerId;
}

#define VFDBC(c, v) c <= v ? v :
#define VFDBCapacity(c) (c <= 2 ? c : VFDBC(c, 4) VFDBC(c, 8) VFDBC(c, 16) VFDBC(c, 32) VFDBC(c, 64) VFDBC(c, 128) VFDBC(c, 256) VFDBC(c, 512) (c / 1024 + (c % 1024 ? 1 : 0) * 1024))
#define VFDBAllCapacity(c) (c <= 2 ? c : VFDBC(c, 4) VFDBC(c, 8) VFDBC(c, 16) VFDBC(c, 32) VFDBC(c, 64) VFDBC(c, 128) VFDBC(c, 256) VFDBC(c, 512) \
	VFDBC(c, 1024) VFDBC(c, 2048) VFDBC(c, 4096) VFDBC(c, 8192) VFDBC(c, 16384) VFDBC(c, 32768) VFDBC(c, 65536) (c / 131072 + (c % 131072 ? 1 : 0) * 131072))

#define VFDBICD(value) cp == value ? value :
#define VFDBIncCapacityDelta(cp) (cp < 2 ? 1 : VFDBICD(2) VFDBICD(4) VFDBICD(8) VFDBICD(16) VFDBICD(32) VFDBICD(64) VFDBICD(128) VFDBICD(256) VFDBICD(512) 1024)
#define VFDBIncAllCapacityDelta(cp) (cp < 2 ? 1 : VFDBICD(2) VFDBICD(4) VFDBICD(8) VFDBICD(16) VFDBICD(32) VFDBICD(64) VFDBICD(128) VFDBICD(256) VFDBICD(512) \
	VFDBICD(1024) VFDBICD(2048) VFDBICD(4096) VFDBICD(8192) VFDBICD(16384) VFDBICD(32768) VFDBICD(65536) 131072)

void CVFDatabase::OpenRecordset(CursorTypeEnum cursorType, bool bind,
	LockTypeEnum lockType, CursorLocationEnum cursorLocation)
{
	CloseRecordset();
	recordset->CursorLocation = cursorLocation;
	recordset->Open((_bstr_t)recordsetName, _variant_t((IDispatch*)connection, true),
		cursorType, lockType, adCmdTableDirect);
	opened = true;
	if(bind)
	{
		TESTHR(recordset->QueryInterface(__uuidof(IADORecordBinding), (LPVOID*)&recordBinding));
		TESTHR(recordBinding->BindToRecordset(&record));
	}
}

void CVFDatabase::CloseRecordset()
{
	if(recordBinding)
	{
		recordBinding->Release();
		recordBinding = NULL;
	}
	if(opened)
	{
		recordset->Close();
		opened = false;
	}
}

void CVFDatabase::Open(const string &provider, const string &dataSource, const string &userId, const string &password,
	IVFDBOpenCallback * callback)
{
	Close();
	string connectionString = Format(TEXT("Provider=%s;Data Source=%s;"), (LPCTSTR)provider, (LPCTSTR)dataSource);

	try
	{
		TESTHR(connection.CreateInstance(__uuidof(Connection)));
		TESTHR(recordset.CreateInstance(__uuidof(Recordset)));

		if(callback) callback->OnOpenState(STRING("Opening connection"));
		connection->Open(connectionString, userId, password, adConnectUnspecified);
		connected = true;
		
		if(callback) callback->OnOpenState(STRING("Retrieving record count"));
		_RecordsetPtr r = connection->Execute(TEXT("SELECT COUNT(*) AS RC FROM ") + recordsetName, NULL, adCmdText);
		int rc = r->Fields->Item["RC"]->Value.lVal;

		if(callback) callback->OnOpenState(STRING("Allocating memory"));
		allCapacity = VFDBAllCapacity(rc);
		allItems = (VFDBItem**)malloc(sizeof(VFDBItem*) * allCapacity);

		if(callback) callback->OnOpenState(STRING("Opening recordset in forward only mode"));
		OpenRecordset(adOpenForwardOnly);
		if(callback)
		{
			callback->OnOpenState(STRING("Loading data"));
			callback->OnOpenProgress(false, rc);
		}
		DWORD dt = GetTickCount();
		while(!recordset->EndOfFile)
		{
			if(callback && GetTickCount() - dt > 300)
			{
				if(!callback->OnOpenProgress(true, allCount)) break;
				dt = GetTickCount();
			}
			if(record.idStatus != adFldOK || record.featuresStatus != adFldOK
				|| record.fingerIdStatus != adFldOK)
				throw exception("Can not open database\nError reading data");
			AddItem(record.id, record.features, record.featuresLength,
				record.fingerId, record.fingerIdLength / sizeof(TCHAR));
			recordset->MoveNext();
		}

		if(callback) callback->OnOpenState(STRING("Opening recordset"));
		OpenRecordset();
		
		if(callback) callback->OnOpenState(STRING("Done"));

		this->provider = provider;
		this->dataSource = dataSource;
	}
	catch(_com_error &e)
	{
		if(callback) callback->OnOpenState(STRING("Error"));
		try
		{
			Close();
		}
		catch(exception &)
		{
			// do nothing
		}
		ComErrorToException(e, STRING("Can not open database"));
	}
}

void CVFDatabase::Close()
{
	ClearItems();
	try
	{
		CloseRecordset();
		recordset = NULL;
		if(connected)
		{
			connection->Close();
			connected = false;
		}
		connection = NULL;
	}
	catch(_com_error &e)
	{
		ComErrorToException(e, STRING("Can not close database"));
	}
	provider.Clear();
	dataSource.Clear();
}

void CVFDatabase::ClearItems()
{
	heap.Clear();
	for(int g = 0; g < 256; g++)
	{
		free(items[g]);
		items[g] = NULL;
		capacity[g] = 0;
		count[g] = 0;
	}
	free(allItems);
	allItems = NULL;
	allCount = 0;
	allCapacity = 0;
}

void CVFDatabase::Clear()
{
	ClearItems();
	if(opened)
	{
		try
		{
			CloseRecordset();
			connection->Execute(Format(TEXT("DELETE FROM %s"), (LPCTSTR)recordsetName),
				NULL, adCmdText | adExecuteNoRecords);
			OpenRecordset();
		}
		catch(_com_error &e)
		{
			ComErrorToException(e, STRING("Can not clear database"));
		}
	}
}

void CVFDatabase::CopyString(LPTSTR dst, DWORD &dstLength, LPCTSTR src, DWORD srcLength, DWORD maxLength)
{
	if(srcLength > maxLength) srcLength = maxLength;
	dstLength = srcLength;
	::CopyString(dst, src, srcLength);
	dst[srcLength] = TEXT('\0');
}

void CVFDatabase::CopyFingerID(LPTSTR dst, DWORD &dstLength, LPCTSTR src, DWORD srcLength)
{
	CopyString(dst, dstLength, src, srcLength, FINGER_ID_LENGTH);
}

int CVFDatabase::AddItem(int id, const BYTE * features, DWORD size,
	LPCTSTR fingerId, DWORD fingerIdLength)
{
	int g = VFFeatGetG(features); VFCheckResult(g);
	int c = count[g];
	int cp = capacity[g];
	if(c == cp)
	{
		cp += VFDBIncCapacityDelta(cp);
		items[g] = (VFDBItem**)realloc(items[g], sizeof(VFDBItem*) * cp);
		capacity[g] = cp;
	}
	if(allCount == allCapacity)
	{
		allCapacity += VFDBIncAllCapacityDelta(allCapacity);
		allItems = (VFDBItem**)realloc(allItems, sizeof(VFDBItem*) * allCapacity);
	}
	VFDBItem *item = (VFDBItem*)heap.Alloc((WORD)sizeof(VFDBItem));
	item->index = allCount;
	item->gIndex = c;
	item->id = id;
	item->features = heap.Alloc((WORD)size);
	item->featuresSize = size;
	memcpy(item->features, features, size);
	CopyFingerID(item->fingerId, item->fingerIdLength, fingerId, fingerIdLength);
	items[g][c] = item;
	allItems[allCount] = item;
	count[g] = ++c;
	return allCount++;
}

void CVFDatabase::SetRecord(int index)
{
	VFDBItem * item = allItems[index];
	memcpy(record.features, item->features, item->featuresSize);
	record.featuresLength = item->featuresSize;
	record.featuresStatus = adFldOK;
	record.fingerIdLength = item->fingerIdLength * sizeof(TCHAR);
	::CopyString(record.fingerId, item->fingerId, item->fingerIdLength);
	record.fingerId[item->fingerIdLength] = TEXT('\0');
	record.fingerIdStatus = adFldOK;
}

bool CVFDatabase::FindRecord(int id)
{
	try
	{
		recordset->MoveFirst();
		recordset->Find(Format(TEXT("ID=%d"), id), 0, adSearchForward);
		return !recordset->EndOfFile;
	}
	catch(_com_error &e)
	{
		ComErrorToException(e, STRING("Can not find record in the database"));
		return false;
	}
}

void CVFDatabase::UpdateRecord(int index)
{
	SetRecord(index);
	FindRecord(allItems[index]->id);
	try
	{
		TESTHR(recordBinding->Update(&record));
	}
	catch(_com_error &e)
	{
		ComErrorToException(e, STRING("Can not update record in the database"));
	}
}

void CVFDatabase::CheckIndex(int index)
{
	if(index < 0 || index >= allCount)
		throw exception("Database record index out of bounds");
}

void CVFDatabase::SetFeatures(int index, const BYTE * features, DWORD size)
{
	CheckIndex(index);
	heap.Free(allItems[index]->features);
	allItems[index]->features = heap.Alloc((WORD)size);
	memcpy(allItems[index]->features, features, size);
	allItems[index]->featuresSize = size;
	if(opened)
	{
		UpdateRecord(index);
	}
}

void CVFDatabase::SetFingerID(int index, const string &fingerId)
{
	CheckIndex(index);
	CopyFingerID(allItems[index]->fingerId, allItems[index]->fingerIdLength, fingerId, fingerId.Length());
	if(opened)
	{
		UpdateRecord(index);
	}
}

int CVFDatabase::Add(const BYTE * features, DWORD size, LPCTSTR fingerId)
{
	int index = AddItem(0, features, size, fingerId, StringLength(fingerId));
	if(opened)
	{
		SetRecord(index);
		try
		{
			TESTHR(recordBinding->AddNew(&record));
			allItems[index]->id = record.id;
		}
		catch(_com_error &e)
		{
			ComErrorToException(e, STRING("Can not add record to the database"));
		}
	}
	return index;
}

void CVFDatabase::Remove(int index)
{
	CheckIndex(index);
	VFDBItem *item = allItems[index];
	int id = item->id;
	int g = VFFeatGetG(item->features); VFCheckResult(g);
	int i =  item->gIndex;
	int c = --count[g];
	memmove(items[g] + i, items[g] + i + 1, sizeof(VFDBItem*) * (c - i));
	int k;
	for(k = i; k < c; k++)
		items[g][k]->gIndex--;
	int cp = VFDBCapacity(c);
	if(cp != capacity[g])
	{
		items[g] = (VFDBItem**)realloc(items[g], sizeof(VFDBItem*) * cp);
		capacity[g] = cp;
	}
	allCount--;
	memmove(allItems + index, allItems + index + 1, sizeof(VFDBItem*) * (allCount - index));
	for(k = index; k < allCount; k++)
		allItems[k]->index--;
	cp = VFDBAllCapacity(allCount);
	if(allCapacity != cp)
	{
		allItems = (VFDBItem**)realloc(allItems, sizeof(VFDBItem*) * cp);
		allCapacity = cp;
	}
	heap.Free(item->features);
	heap.Free((BYTE*)item);
	if(opened)
	{
		if(FindRecord(id))
			try
			{
				recordset->Delete(adAffectCurrent);
			}
			catch(_com_error &e)
			{
				ComErrorToException(e, STRING("Can not delete record from the database"));
			}
	}
}

void CVFDatabase::FindStart(const BYTE * features, VFDBFindData &fd)
{
	fd.g = fd.testG = VFFeatGetG(features); VFCheckResult(fd.testG);
	fd.index = -1;
}

int CVFDatabase::FindNext(VFDBFindData &fd)
{
	fd.index++;
	while(fd.index >= count[fd.g])
	{
		fd.index = 0;
		int gDif = fd.g - fd.testG;
		fd.g = fd.testG + (gDif == 0 ? 1 : gDif > 0 ? -gDif : -gDif + 1);
		gDif = fd.g - fd.testG;
		if(fd.g >= 255)
		{
			fd.g = fd.testG - gDif;
			if(fd.g < 0) return -1;
		}
		else if(fd.g < 0)
		{
			fd.g = fd.testG - gDif + 1;
			if(fd.g >= 255) return -1;
		}
	}
	return items[fd.g][fd.index]->index;
}

void CVFDatabase::FindEnd(VFDBFindData &/*fd*/)
{
}

const string defaultDatabaseProvider(TEXT("Microsoft.Jet.OLEDB.4.0"));
const string defaultDatabaseFileExtension(TEXT("mdb"));

string GetVFDBDefaultFilter()
{
	return TEXT("*.") + defaultDatabaseFileExtension;
}

string GetVFDBDefaultFilterName()
{
	return TEXT("Microsoft Access Databases");
}

string GetVFDBDefaultFilterString(bool filterInName)
{
	return MakeFileFilterString(GetVFDBDefaultFilterName(), GetVFDBDefaultFilter(), filterInName);
}
