﻿using System;
using System.Collections.Generic;
using System.Text;

namespace Neurotec.Biometrics.VeriFingerDemo
{
	public class GIDistribution
	{
		public GIDistribution(Distribution genuineDistribution, Distribution impostorDistribution, bool reversePolarity)
		{
			if(genuineDistribution == null) throw new ArgumentNullException("genuineDistribution");
			if(impostorDistribution == null) throw new ArgumentNullException("impostorDistribution");
			genuines = genuineDistribution;
			impostors = impostorDistribution;
			this.reversePolarity = reversePolarity;
		}

		public GIDistribution(IList<int> genuineItems, int genuinesMin, IList<int> impostorsItems, int impostorsMin, bool reversePolarity)
			: this(new Distribution(genuineItems, genuinesMin), new Distribution(impostorsItems, impostorsMin), reversePolarity)
		{
		}

		private bool reversePolarity;

		public bool ReversePolarity
		{
			get
			{
				return reversePolarity;
			}
		}

		private Distribution genuines;

		public Distribution Genuines
		{
			get
			{
				return genuines;
			}
		}

		private Distribution impostors;

		public Distribution Impostors
		{
			get
			{
				return impostors;
			}
		}

		public int ValueCount
		{
			get
			{
				return genuines.ValueCount + impostors.ValueCount;
			}
		}

		public int CommonMin
		{
			get
			{
				return Math.Max(genuines.Min, impostors.Min);
			}
		}

		public int CommonMax
		{
			get
			{
				return Math.Min(genuines.Max, impostors.Max);
			}
		}

		public bool IsEmpty
		{
			get
			{
				return genuines.IsEmpty || impostors.IsEmpty;
			}
		}

		public int GetFac(int value)
		{
			return reversePolarity ? impostors.GetLTECount(value) : impostors.GetGTECount(value);
		}

		public int GetFrc(int value)
		{
			return reversePolarity ? genuines.GetGTCount(value) : genuines.GetLTCount(value);
		}

		public int GetTrc(int value)
		{
			return reversePolarity ? impostors.GetGTCount(value) : impostors.GetLTCount(value);
		}

		public int GetTac(int value)
		{
			return reversePolarity ? genuines.GetLTECount(value) : genuines.GetGTECount(value);
		}

		public double GetFar(int value)
		{
			return impostors.CountToProbability(GetFac(value));
		}

		public double GetFrr(int value)
		{
			return genuines.CountToProbability(GetFrc(value));
		}

		public double GetTrr(int value)
		{
			return impostors.CountToProbability(GetTrc(value));
		}

		public double GetTar(int value)
		{
			return genuines.CountToProbability(GetTac(value));
		}

		private static void CheckPercent(double value)
		{
			if(value < 0 || value > 1) throw new ArgumentOutOfRangeException("value", "value is less than zero or greater than one");
		}

		public int GetValueAtFar(double far)
		{
			CheckPercent(far);
			int min = impostors.Min;
			int max = impostors.Max;
			int s, e, d;
			if(reversePolarity)
			{
				s = max;
				e = min - 1;
				d = -1;
			}
			else
			{
				s = min;
				e = max + 1;
				d = 1;
			}
			double lastFar = GetFar(s);
			double lastDelta = far - lastFar;
			int value = s;
			for(++value; value != e; value += d)
			{
				double curFar = GetFar(value);
				double curDelta = far - curFar;
				if(lastDelta * curDelta <= 0)
				{
					if(lastDelta != 0) value += d;
					break;
				}
				lastFar = curFar;
				lastDelta = curDelta;
			}
			return value - d;
		}

		public int GetValueAtFrr(double frr)
		{
			CheckPercent(frr);
			int min = genuines.Min;
			int max = genuines.Max;
			int s, e, d;
			if(reversePolarity)
			{
				s = min;
				e = max + 1;
				d = 1;
			}
			else
			{
				s = max;
				e = min - 1;
				d = -1;
			}
			double lastFrr = GetFrr(s);
			double lastDelta = frr - lastFrr;
			int value = s;
			for(++value; value != e; value += d)
			{
				double curFrr = GetFrr(value);
				double curDelta = frr - curFrr;
				if(lastDelta * curDelta <= 0)
				{
					if(lastDelta != 0) value += d;
					break;
				}
				lastFrr = curFrr;
				lastDelta = curDelta;
			}
			return value - d;
		}

		public double GetValueAtTrr(double trr)
		{
			return GetValueAtFar(1 - trr);
		}

		public double GetValueAtTar(double tar)
		{
			return GetValueAtFrr(1 - tar);
		}

		public double GetFarAtFrr(double frr)
		{
			return GetFar(GetValueAtFrr(frr));
		}

		public double GetFrrAtFar(double far)
		{
			return GetFrr(GetValueAtFar(far));
		}

		public double GetFarAtTar(double tar)
		{
			return GetFarAtFrr(1 - tar);
		}

		public double GetTarAtFar(double far)
		{
			return 1 - GetFrrAtFar(far);
		}

		public double GetTrrAtFrr(double frr)
		{
			return 1 - GetFarAtFrr(frr);
		}

		public double GetFrrAtTrr(double trr)
		{
			return GetFrrAtFar(1 - trr);
		}

		public double GetTrrAtTar(double tar)
		{
			return 1 - GetFarAtFrr(1 - tar);
		}

		public double GetTarAtTrr(double trr)
		{
			return 1 - GetFrrAtFar(1 - trr);
		}

		public double GetFarFrrRatio(int value)
		{
			return GetFar(value) / GetFrr(value);
		}

		public int GetValueAtFarFrrRatio(double farFrrRatio)
		{
			if(IsEmpty) return 0;
			else if(genuines.IsEmpty) return reversePolarity ? impostors.Max : impostors.Min;
			else if(impostors.IsEmpty) return reversePolarity ? genuines.Max : genuines.Min;
			int min = CommonMin;
			int max = CommonMax;
			if(min > max) return (min + max + (reversePolarity ? 0 : 1)) / 2;
			int s, e, d;
			if(reversePolarity)
			{
				s = max;
				e = min - 1;
				d = -1;
			}
			else
			{
				s = min;
				e = max + 1;
				d = 1;
			}
			double lastFar = GetFar(s);
			double lastFrr = GetFrr(s);
			double lastDelta = farFrrRatio - lastFar / lastFrr;
			int value = s;
			for(++value; value != e; value += d)
			{
				double curFar = GetFar(value);
				double curFrr = GetFrr(value);
				double curDelta = farFrrRatio - curFar / curFrr;
				if(lastDelta * curDelta <= 0)
				{
					if(lastDelta != 0) value += d;
					break;
				}
				lastFar = curFar;
				lastFrr = curFrr;
				lastDelta = curDelta;
			}
			return value - d;
		}

		public int GetValueAtEer()
		{
			return GetValueAtFarFrrRatio(1);
		}

		public double GetEer()
		{
			int value = GetValueAtEer();
			return (GetFar(value) + GetFrr(value)) / 2;
		}

		public double GetWer(int value)
		{
			double frr = GetFrr(value);
			double far = GetFar(value);
			return (frr * frr + far * far) / (frr + far);
		}

		public double GetWer(int value, double farFrrRatio)
		{
			double frr = GetFrr(value);
			double far = GetFar(value);
			return (frr + farFrrRatio * far) / (1 + farFrrRatio);
		}

		public double GetWer(int value, GIDistribution priorGIDistribution)
		{
			if(priorGIDistribution == null) throw new ArgumentNullException("priorGIDistribution");
			return GetWer(value, priorGIDistribution.GetFarFrrRatio(value));
		}

		public double GetWerForRatio(double farFrrRatio)
		{
			return GetFrr(GetValueAtFarFrrRatio(farFrrRatio)) * (1 + farFrrRatio * farFrrRatio) / (1 + farFrrRatio);
		}

		public IEnumerable<CountValuePair> GetFacPoints()
		{
			if (impostors.IsEmpty) yield break;
			int min = impostors.Min;
			int max = impostors.Max;
			int s, e, d;
			if(reversePolarity)
			{
				s = max;
				e = min - 1;
				d = -1;
			}
			else
			{
				s = min;
				e = max + 1;
				d = 1;
			}
			int lastFac = 0;
			bool lastAdded = true;
			for(int value = s; value != e + d; value += d)
			{
				int curFac = GetFac(value);
				if(value == s || value == e || curFac != lastFac)
				{
					if(!lastAdded) yield return new CountValuePair(lastFac, value - d);
					yield return new CountValuePair(curFac, value);
					lastAdded = true;
				}
				else lastAdded = false;
				lastFac = curFac;
			}
		}

		public IEnumerable<CountValuePair> GetFrcPoints()
		{
			if(genuines.IsEmpty) yield break;
			int min = genuines.Min;
			int max = genuines.Max;
			int s, e, d;
			if(reversePolarity)
			{
				s = max;
				e = min - 1;
				d = -1;
			}
			else
			{
				s = min;
				e = max + 1;
				d = 1;
			}
			int lastFrc = 0;
			bool lastAdded = true;
			for(int value = s; value != e + d; value += d)
			{
				int curFrc = GetFrc(value);
				if(value == s || value == e || curFrc != lastFrc)
				{
					if(!lastAdded) yield return new CountValuePair(lastFrc, value - d);
					yield return new CountValuePair(curFrc, value);
					lastAdded = true;
				}
				else lastAdded = false;
				lastFrc = curFrc;
			}
		}

		public IEnumerable<CountValuePair> GetTrcPoints()
		{
			int count = impostors.ValueCount;
			foreach (CountValuePair point in GetFacPoints())
			{
				yield return new CountValuePair(count - point.Count, point.Value);
			}
		}

		public IEnumerable<CountValuePair> GetTacPoints()
		{
			int count = genuines.ValueCount;
			foreach (CountValuePair point in GetFrcPoints())
			{
				yield return new CountValuePair(count - point.Count, point.Value);
			}
		}

		public IEnumerable<ProbabilityValuePair> GetFarPoints()
		{
			double count = (double)impostors.ValueCount;
			foreach (CountValuePair point in GetFacPoints())
			{
				yield return new ProbabilityValuePair(point.Count / count, point.Value);
			}
		}

		public IEnumerable<ProbabilityValuePair> GetFrrPoints()
		{
			double count = (double)genuines.ValueCount;
			foreach(CountValuePair point in GetFrcPoints())
			{
				yield return new ProbabilityValuePair(point.Count / count, point.Value);
			}
		}

		public IEnumerable<ProbabilityValuePair> GetTrrPoints()
		{
			double count = (double)impostors.ValueCount;
			foreach (CountValuePair point in GetTrcPoints())
			{
				yield return new ProbabilityValuePair(point.Count / count, point.Value);
			}
		}

		public IEnumerable<ProbabilityValuePair> GetTarPoints()
		{
			double count = (double)genuines.ValueCount;
			foreach (CountValuePair point in GetTacPoints())
			{
				yield return new ProbabilityValuePair(point.Count / count, point.Value);
			}
		}

		public IEnumerable<ProbabilityValuePair> GetWerPoints()
		{
			if(IsEmpty) yield break;
			int min = CommonMin;
			int max = CommonMax;
			if(max < min)
			{
				int valueAtZeroFar = GetValueAtFar(0);
				int valueAtEer = GetValueAtEer();
				int valueAtZeroFrr = GetValueAtFrr(0);
				yield return new ProbabilityValuePair(GetWer(valueAtZeroFar), valueAtZeroFar);
				yield return new ProbabilityValuePair(GetWer(valueAtEer), valueAtEer);
				yield return new ProbabilityValuePair(GetWer(valueAtZeroFrr), valueAtZeroFrr);
			}
			int s, e, d;
			if(reversePolarity)
			{
				s = max;
				e = min - 1;
				d = -1;
			}
			else
			{
				s = min;
				e = max + 1;
				d = 1;
			}
			double lastWer = 0;
			bool lastAdded = true;
			for(int value = s; value != e + d; value += d)
			{
				double curWer = GetWer(value);
				if(value == s || value == e || curWer != lastWer)
				{
					if(!lastAdded) yield return new ProbabilityValuePair(lastWer, value - d);
					yield return new ProbabilityValuePair(curWer, value);
					lastAdded = true;
				}
				else lastAdded = false;
				lastWer = curWer;
			}
		}

		public IEnumerable<ProbabilityValuePair> GetWerPoints(GIDistribution priorGIDistribution)
		{
			if(priorGIDistribution == null) throw new ArgumentNullException("priorGIDistribution");
			if (IsEmpty) yield break;
			int min = CommonMin;
			int max = CommonMax;
			if(max < min)
			{
				int valueAtZeroFar = GetValueAtFar(0);
				int valueAtEer = GetValueAtEer();
				int valueAtZeroFrr = GetValueAtFrr(0);
				yield return new ProbabilityValuePair(GetWer(valueAtZeroFar), valueAtZeroFar);
				yield return new ProbabilityValuePair(GetWer(valueAtEer), valueAtEer);
				yield return new ProbabilityValuePair(GetWer(valueAtZeroFrr), valueAtZeroFrr);
			}
			int s, e, d;
			if(reversePolarity)
			{
				s = max;
				e = min - 1;
				d = -1;
			}
			else
			{
				s = min;
				e = max + 1;
				d = 1;
			}
			double lastWer = 0;
			bool lastAdded = true;
			for(int value = s; value != e + d; value += d)
			{
				double curWer = GetWer(value, priorGIDistribution.GetFarFrrRatio(value));
				if(value == s || value == e || curWer != lastWer)
				{
					if(!lastAdded) yield return new ProbabilityValuePair(lastWer, value - d);
					yield return new ProbabilityValuePair(curWer, value);
					lastAdded = true;
				}
				else lastAdded = false;
				lastWer = curWer;
			}
		}

		public IEnumerable<GICountValueTriple> GetFacAgainstFrcPoints()
		{
			if (IsEmpty) yield break;
			int min = CommonMin;
			int max = CommonMax;
			if(max < min)
			{
				int valueAtZeroFar = GetValueAtFar(0);
				int valueAtEer = GetValueAtEer();
				int valueAtZeroFrr = GetValueAtFrr(0);
				yield return new GICountValueTriple(GetFrc(valueAtZeroFar), GetFac(valueAtZeroFar), valueAtZeroFar);
				yield return new GICountValueTriple(GetFrc(valueAtEer), GetFac(valueAtEer), valueAtEer);
				yield return new GICountValueTriple(GetFrc(valueAtZeroFrr), GetFac(valueAtZeroFrr), valueAtZeroFrr);
			}
			List<GICountValueTriple> points = new List<GICountValueTriple>();
			int s, e, d;
			if(reversePolarity)
			{
				s = max;
				e = min - 1;
				d = -1;
			}
			else
			{
				s = min;
				e = max + 1;
				d = 1;
			}
			int lastFac = 0;
			int lastFrc = 0;
			bool lastAdded = true;
			for(int value = s; value != e + d; value += d)
			{
				int curFac = GetFac(value);
				int curFrc = GetFrc(value);
				if(value == s || value == e || curFac != lastFac || curFrc != lastFrc)
				{
					if(!lastAdded) yield return new GICountValueTriple(lastFrc, lastFac, value - d);
					yield return new GICountValueTriple(curFrc, curFac, value);
					lastAdded = true;
				}
				else lastAdded = false;
				lastFac = curFac;
				lastFrc = curFrc;
			}
		}

		public IEnumerable<GICountValueTriple> GetFacAgainstTacPoints()
		{
			int count = genuines.ValueCount;
			foreach(GICountValueTriple point in GetFacAgainstFrcPoints())
			{
				yield return new GICountValueTriple(count - point.GenuineCount, point.ImpostorCount, point.Value);
			}
		}

		public IEnumerable<GICountValueTriple> GetTrcAgainstFrcPoints()
		{
			int count = impostors.ValueCount;
			foreach (GICountValueTriple point in GetFacAgainstFrcPoints())
			{
				yield return new GICountValueTriple(point.GenuineCount, count - point.ImpostorCount, point.Value);
			}
		}

		public IEnumerable<GICountValueTriple> GetTrcAgainstTacPoints()
		{
			int gCount = genuines.ValueCount;
			int iCount = impostors.ValueCount;
			foreach (GICountValueTriple point in GetFacAgainstFrcPoints())
			{
				yield return new GICountValueTriple(gCount - point.GenuineCount, iCount - point.ImpostorCount, point.Value);
			}
		}

		public IEnumerable<GIProbabilityValueTriple> GetFarAgainstFrrPoints()
		{
			foreach(GICountValueTriple point in GetFacAgainstFrcPoints())
			{
				yield return new GIProbabilityValueTriple(
					point.GenuineCount / (double)genuines.ValueCount,
					point.ImpostorCount / (double)impostors.ValueCount,
					point.Value);
			}
		}

		public IEnumerable<GIProbabilityValueTriple> GetFarAgainstTarPoints()
		{
			foreach(GICountValueTriple point in GetFacAgainstTacPoints())
			{
				yield return new GIProbabilityValueTriple(
					point.GenuineCount / (double)genuines.ValueCount,
					point.ImpostorCount / (double)impostors.ValueCount,
					point.Value);
			}
		}

		public IEnumerable<GIProbabilityValueTriple> GetTrrAgainstFrrPoints()
		{
			foreach(GICountValueTriple point in GetTrcAgainstFrcPoints())
			{
				yield return new GIProbabilityValueTriple(
					point.GenuineCount / (double)genuines.ValueCount,
					point.ImpostorCount / (double)impostors.ValueCount,
					point.Value);
			}
		}

		public IEnumerable<GIProbabilityValueTriple> GetTrrAgainstTarPoints()
		{
			foreach(GICountValueTriple point in GetTrcAgainstTacPoints())
			{
				yield return new GIProbabilityValueTriple(
					point.GenuineCount / (double)genuines.ValueCount,
					point.ImpostorCount / (double)impostors.ValueCount,
					point.Value);
			}
		}
	}

	public struct GICountValueTriple
	{
		public GICountValueTriple(int genuineCount, int impostorCount, int value)
		{
			GenuineCount = genuineCount;
			ImpostorCount = impostorCount;
			Value = value;
		}

		public int GenuineCount;
		public int ImpostorCount;
		public int Value;
	}

	public struct GIProbabilityValueTriple
	{
		public GIProbabilityValueTriple(double genuineProbability, double impostorProbability, int value)
		{
			GenuineProbability = genuineProbability;
			ImpostorProbability = impostorProbability;
			Value = value;
		}

		public double GenuineProbability;
		public double ImpostorProbability;
		public int Value;
	}
}
